#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python KISS Module Constants."""

import logging
import os

__author__ = 'Greg Albrecht W2GMD <oss@undef.net>'  # NOQA pylint: disable=R0801
__copyright__ = 'Copyright 2017 Greg Albrecht and Contributors'  # NOQA pylint: disable=R0801
__license__ = 'Apache License, Version 2.0'  # NOQA pylint: disable=R0801


if bool(os.environ.get('DEBUG')):
    LOG_LEVEL = logging.DEBUG
    logging.debug('Debugging Enabled via DEBUG Environment Variable.')
else:
    LOG_LEVEL = logging.INFO

LOG_FORMAT = logging.Formatter(
    '%(asctime)s kiss %(levelname)s %(name)s.%(funcName)s:%(lineno)d'
    ' - %(message)s')

SERIAL_TIMEOUT = 0.01
READ_BYTES = 1000

# KISS Special Characters
# http://en.wikipedia.org/wiki/KISS_(TNC)#Special_Characters
# http://k4kpk.com/content/notes-aprs-kiss-and-setting-tnc-x-igate-and-digipeater
# Frames begin and end with a FEND/Frame End/0xC0 byte
FEND = b'\xC0'  # Marks START and END of a Frame
FESC = b'\xDB'  # Escapes FEND and FESC bytes within a frame

# Transpose Bytes: Used within a frame-
# "Transpose FEND": An FEND after an FESC (within a frame)-
# Sent as FESC TFEND
TFEND = b'\xDC'
# "Transpose FESC": An FESC after an FESC (within a frame)-
# Sent as FESC TFESC
TFESC = b'\xDD'

# "FEND is sent as FESC, TFEND"
# 0xC0 is sent as 0xDB 0xDC
FESC_TFEND = b''.join([FESC, TFEND])

# "FESC is sent as FESC, TFESC"
# 0xDB is sent as 0xDB 0xDD
FESC_TFESC = b''.join([FESC, TFESC])

# KISS Command Codes
# http://en.wikipedia.org/wiki/KISS_(TNC)#Command_Codes
DATA_FRAME = b'\x00'
TX_DELAY = b'\x01'
PERSISTENCE = b'\x02'
SLOT_TIME = b'\x03'
TX_TAIL = b'\x04'
FULL_DUPLEX = b'\x05'
SET_HARDWARE = b'\x06'
RETURN = b'\xFF'

# Alternate convenience spellings for Command Codes
# (these more closely match the names in the spec)
DATAFRAME = DATA_FRAME
TXDELAY = TX_DELAY
P = PERSISTENCE
SLOTTIME = SLOT_TIME
TXTAIL = TX_TAIL
FULLDUPLEX = FULL_DUPLEX
SETHARDWARE = SET_HARDWARE

DEFAULT_KISS_CONFIG_VALUES = {
    'TX_DELAY': 40,
    'PERSISTENCE': 63,
    'SLOT_TIME': 20,
    'TX_TAIL': 30,
    'FULL_DUPLEX': 0,
}

KISS_ON = 'KISS $0B'
KISS_OFF = b''.join([FEND, RETURN, FEND, FEND])

NMEA_HEADER = b''.join([FEND, b'\xF0', b'$'])

UI_PROTOCOL_ID = b'\xF0'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python KISS Module Utility Functions Definitions."""

from . import constants

__author__ = 'Greg Albrecht W2GMD <oss@undef.net>'  # NOQA pylint: disable=R0801
__copyright__ = 'Copyright 2017 Greg Albrecht and Contributors'  # NOQA pylint: disable=R0801
__license__ = 'Apache License, Version 2.0'  # NOQA pylint: disable=R0801


def escape_special_codes(raw_codes):
    """
    Escape special codes, per KISS spec.

    "If the FEND or FESC codes appear in the data to be transferred, they
    need to be escaped. The FEND code is then sent as FESC, TFEND and the
    FESC is then sent as FESC, TFESC."
    - http://en.wikipedia.org/wiki/KISS_(TNC)#Description
    """
    output = b''
    for i, x in enumerate(raw_codes):
        x = bytes([x])
        if constants.FESC == x:
            output += constants.FESC_TFESC
        elif constants.FEND == x:
            output += constants.FESC_TFEND
        else:
            output += x
    return output


def recover_special_codes(escaped_codes):
    """
    Recover special codes, per KISS spec.

    "If the FESC_TFESC or FESC_TFEND escaped codes appear in the data received,
    they need to be recovered to the original codes. The FESC_TFESC code is
    replaced by FESC code and FESC_TFEND is replaced by FEND code."
    - http://en.wikipedia.org/wiki/KISS_(TNC)#Description
    """

    output = b''
    enumerated = enumerate(escaped_codes)
    for i, x in enumerated:
        if constants.FESC_TFESC == escaped_codes[i:i+2]:
            output += constants.FESC
            next(enumerated, None)
        elif constants.FESC_TFEND == escaped_codes[i:i+2]:
            output += constants.FEND
            next(enumerated, None)
        else:
            output += bytes([x])
    return output

def extract_ui(frame):
    """
    Extracts the UI component of an individual frame.

    :param frame: APRS/AX.25 frame.
    :type frame: str
    :returns: UI component of frame.
    :rtype: str
    """
    start_ui = frame.split(
        b''.join([constants.FEND, constants.DATA_FRAME]))
    end_ui = start_ui[0].split(b''.join([constants.SLOT_TIME, constants.UI_PROTOCOL_ID]))
    return ''.join([chr(x >> 1) for x in end_ui[0]])


def strip_df_start(frame):
    """
    Strips KISS DATA_FRAME start (0x00) and newline from frame.

    :param frame: APRS/AX.25 frame.
    :type frame: str
    :returns: APRS/AX.25 frame sans DATA_FRAME start (0x00).
    :rtype: str
    """
    return frame.lstrip(constants.DATA_FRAME).strip()


def strip_nmea(frame):
    """
    Extracts NMEA header from T3-Micro or NMEA encoded KISS frames.
    """
    if len(frame) > 0:
        if frame[0] == 240:
            return frame[1:].rstrip()
    return frame



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python KISS Module Exception Definitions."""

__author__ = 'Greg Albrecht W2GMD <oss@undef.net>'  # NOQA pylint: disable=R0801
__copyright__ = 'Copyright 2017 Greg Albrecht and Contributors'  # NOQA pylint: disable=R0801
__license__ = 'Apache License, Version 2.0'  # NOQA pylint: disable=R0801


class SocketClosetError(Exception):
    """Socket Closed Error."""
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python KISS Module Class Definitions."""

import logging
import socket

import serial

from . import constants, util

__author__ = 'Greg Albrecht W2GMD <oss@undef.net>'  # NOQA pylint: disable=R0801
__copyright__ = 'Copyright 2017 Greg Albrecht and Contributors'  # NOQA pylint: disable=R0801
__license__ = 'Apache License, Version 2.0'  # NOQA pylint: disable=R0801


class KISS(object):

    """KISS Object Class."""

    _logger = logging.getLogger(__name__)  # pylint: disable=R0801
    if not _logger.handlers:  # pylint: disable=R0801
        _logger.setLevel(constants.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler = logging.StreamHandler()  # pylint: disable=R0801
        _console_handler.setLevel(constants.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler.setFormatter(constants.LOG_FORMAT)  # pylint: disable=R0801
        _logger.addHandler(_console_handler)  # pylint: disable=R0801
        _logger.propagate = False  # pylint: disable=R0801

    def __init__(self, strip_df_start: bool=False) -> None:
        self.strip_df_start = strip_df_start
        self.interface = None
        self.reading = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def __del__(self):
        self.stop()

    def _read_handler(self, read_bytes=None):  # pylint: disable=R0201
        """
        Helper method to call when reading from KISS interface.
        """
        read_bytes = read_bytes or constants.READ_BYTES

    def _write_handler(self, frame=None):  # pylint: disable=R0201
        """
        Helper method to call when writing to KISS interface.
        """
        del frame

    def stop(self):
        """
        Helper method to call when stopping KISS interface.
        """
        pass

    def start(self, **kwargs):
        """
        Helper method to call when starting KISS interface.
        """
        pass

    def write_setting(self, name, value):
        """
        Writes KISS Command Codes to attached device.

        http://en.wikipedia.org/wiki/KISS_(TNC)#Command_Codes

        :param name: KISS Command Code Name as a string.
        :param value: KISS Command Code Value to write.
        """
        self._logger.debug('Configuring %s=%s', name, repr(value))

        # Do the reasonable thing if a user passes an int
        if isinstance(value, int):
            value = chr(value)

        return self.interface.write(
            b''.join([
                constants.FEND,
                bytes(getattr(kiss, name.upper())),
                constants.escape_special_codes(value),
                constants.FEND
            ])
        )

    def read(self, read_bytes=None, callback=None, readmode=True):  # NOQA pylint: disable=R0912
        """
        Reads data from KISS device.

        :param callback: Callback to call with decoded data.
        :param readmode: If False, immediately returns frames.
        :type callback: func
        :type readmode: bool
        :return: List of frames (if readmode=False).
        :rtype: list
        """
        self._logger.debug(
            'read_bytes=%s callback="%s" readmode=%s',
            read_bytes, callback, readmode)

        read_buffer = bytes()

        self.reading = True

        while self.reading == True:
            read_data = self._read_handler(read_bytes)

            if read_data is not None and len(read_data):
                self._logger.debug(
                    'read_data(%s)="%s"', len(read_data), read_data)

                frames = []

                split_data = read_data.split(constants.FEND)
                fends = len(split_data)

                self._logger.debug(
                    'split_data(fends=%s)="%s"', fends, split_data)

                # Handle NMEAPASS on T3-Micro
                if len(read_data) >= 900:
                    if constants.NMEA_HEADER in read_data and '\r\n' in read_data:
                        if callback:
                            callback(read_data)
                        elif not readmode:
                            return [read_data]

                # No FEND in frame
                if fends == 1:
                    read_buffer += split_data[0]
                # Single FEND in frame
                elif fends == 2:
                    # Closing FEND found
                    if split_data[0]:
                        # Partial frame continued, otherwise drop
                        frames.append(b''.join([read_buffer, split_data[0]]))
                        read_buffer = bytes()
                    # Opening FEND found
                    else:
                        frames.append(read_buffer)
                        read_buffer = split_data[1]

                # At least one complete frame received: [FEND, xxx, FEND]
                elif fends >= 3:

                    # Iterate through split_data and extract just the frames.
                    for i in range(0, fends - 1):
                        buf = bytearray(b''.join([read_buffer, split_data[i]]))
                        self._logger.debug('i=%s buf="%s"', i, buf)
                        if buf:
                            self._logger.debug('Frame Found: "%s"', buf)
                            frames.append(buf)
                            read_buffer = bytearray()

                    # TODO: What do I do?
                    if split_data[fends - 1]:
                        self._logger.debug('Mystery Conditional')
                        read_buffer = bytearray(split_data[fends - 1])

                # Fixup T3-Micro NMEA Sentences
                frames = list(map(util.strip_nmea, frames))
                # Remove None frames.
                frames = [_f for _f in frames if _f]

                # Maybe.
                frames = list(map(util.recover_special_codes, frames))

                if self.strip_df_start:
                    frames = list(map(util.strip_df_start, frames))

                if readmode:
                    for frame in frames:
                        callback(frame)
                elif not readmode:
                    return frames

    def write(self, frame):
        """
        Writes frame to KISS interface.

        :param frame: Frame to write.
        """
        self._logger.debug('frame(%s)="%s"', len(frame), frame)

        frame_escaped = util.escape_special_codes(frame)
        self._logger.debug(
            'frame_escaped(%s)="%s"', len(frame_escaped), frame_escaped)

        frame_kiss = constants.FEND + constants.DATA_FRAME + frame_escaped + constants.FEND
        self._logger.debug(
            'frame_kiss(%s)="%s"', len(frame_kiss), frame_kiss)

        self._write_handler(frame_kiss)


class TCPKISS(KISS):

    """KISS TCP Class."""

    def __init__(self, host, port, strip_df_start=False):
        self.address = (host, int(port))
        self.strip_df_start = strip_df_start
        super(TCPKISS, self).__init__(strip_df_start)

    def _read_handler(self, read_bytes=None):
        read_bytes = read_bytes or constants.READ_BYTES
        read_data = self.interface.recv(read_bytes)
        self._logger.debug('len(read_data)=%s', len(read_data))
        return read_data

    def stop(self):
        self.reading = False
        if self.interface:
            self.interface.shutdown(socket.SHUT_RDWR)

    def start(self):
        """
        Initializes the KISS device and commits configuration.
        """
        self.interface = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._logger.debug('Conntecting to %s', self.address)
        self.interface.connect(self.address)
        self._logger.info('Connected to %s', self.address)
        self._write_handler = self.interface.send

class TCPServerKISS(KISS):

    """KISS TCP Class."""

    def __init__(self, host, port, strip_df_start=False):
        self.address = (host, int(port))
        self.strip_df_start = strip_df_start
        self.conns = []
        super(TCPServerKISS, self).__init__(strip_df_start)
    
    def _write_handler(self, frame):
        conns = self.conns
        try:
            conn, addr = self.interface.accept()
            self._logger.info(f"Client connected from {addr}")
            self.conns.append(conn)
        except socket.timeout:
            pass
        conns = self.conns
        for conn in conns:
            try:
                conn.send(frame)
            except ConnectionResetError:
                self.conns.remove(conn)
            except BrokenPipeError:
                self.conns.remove(conn)

    def _read_handler(self, read_bytes=None):
        read_bytes = read_bytes or constants.READ_BYTES
        try:
            conn, addr = self.interface.accept()
            self._logger.info(f"Client connected from {addr}")
            conn.settimeout(0.01)
            self.conns.append(conn)
        except socket.timeout:
            pass
        conns = self.conns
        for conn in self.conns:
            try: 
                read_data = conn.recv(read_bytes)
                if read_data:
                    self._logger.info(f"Read KISS bytes from TCP client : {read_bytes}")
                    self._logger.debug('len(read_data)=%s', len(read_data))
                    return read_data
            except socket.timeout:
                pass
            except ConnectionResetError: 
                self._logger.info(f"{conn} Disconnected")
                self.conns.remove(conn)

    def stop(self):
        self.reading = False
        if self.interface:
            self.interface.shutdown(socket.SHUT_RDWR)

    def start(self):
        """
        Initializes the KISS device and commits configuration.
        """
        self.interface = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.interface.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.interface.bind(self.address)
        self.interface.listen(1)
        self.interface.settimeout(0.01)
        self._logger.info('Server started')
        #self._write_handler = self.interface.send

class SerialKISS(KISS):

    """KISS Serial Class."""

    def __init__(self, port: str, speed: str,
                 strip_df_start: bool=False) -> None:
        self.port = port
        self.speed = speed
        self.strip_df_start = strip_df_start
        super(SerialKISS, self).__init__(strip_df_start)

    def _read_handler(self, read_bytes=None):
        read_bytes = read_bytes or constants.READ_BYTES
        read_data = self.interface.read(read_bytes)
        if len(read_data) > 0:
            self._logger.debug('len(read_data)=%s', len(read_data))

        try:
            waiting_data = self.interface.in_waiting
        except AttributeError:
            waiting_data = self.interface.outWaiting()

        if waiting_data:
            self._logger.debug('waiting_data=%s', waiting_data)
            read_data += self.interface.read(waiting_data)
        return read_data

    def _write_defaults(self, **kwargs):
        """
        Previous verious defaulted to Xastir-friendly configs. Unfortunately
        those don't work with Bluetooth TNCs, so we're reverting to None.

        Use `config_xastir()` for Xastir defaults.
        """
        return [self.write_setting(k, v) for k, v in list(kwargs.items())]

    def config_xastir(self):
        """
        Helper method to set default configuration to those that ship with
        Xastir.
        """
        return self._write_defaults(
            **constants.DEFAULT_KISS_CONFIG_VALUES)

    def kiss_on(self):
        """Turns KISS ON."""
        self.interface.write(constants.KISS_ON)

    def kiss_off(self):
        """Turns KISS OFF."""
        self.interface.write(constants.KISS_OFF)

    def stop(self):
        self.reading = False
        try:
            if self.interface and self.interface.isOpen():
                self.interface.close()
        except AttributeError:
            if self.interface and self.interface._isOpen:
                self.interface.close()

    def start(self, **kwargs):
        """
        Initializes the KISS device and commits configuration.

        See http://en.wikipedia.org/wiki/KISS_(TNC)#Command_codes
        for configuration names.

        :param **kwargs: name/value pairs to use as initial config values.
        """
        self._logger.debug('kwargs=%s', kwargs)
        self.interface = serial.Serial(self.port, self.speed)
        self.interface.timeout = constants.SERIAL_TIMEOUT
        self._write_handler = self.interface.write
        self._write_defaults(**kwargs)

    def start_no_config(self):
        """
        Initializes the KISS device without writing configuration.
        """
        self.interface = serial.Serial(self.port, self.speed)
        self.interface.timeout = constants.SERIAL_TIMEOUT
        self._write_handler = self.interface.write



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Python KISS Module.

"""
Python KISS Module.
~~~~


:author: Greg Albrecht W2GMD <oss@undef.net>
:copyright: Copyright 2017 Greg Albrecht and Contributors
:license: Apache License, Version 2.0
:source: <https://github.com/ampledata/kiss>

"""

from .constants import (LOG_FORMAT, LOG_LEVEL, SERIAL_TIMEOUT, READ_BYTES,  # NOQA
                        FEND, FESC, TFEND, TFESC, FESC_TFEND, FESC_TFESC,
                        DATA_FRAME, TX_DELAY, PERSISTENCE, SLOT_TIME, TX_TAIL,
                        FULL_DUPLEX, SET_HARDWARE, RETURN, DATAFRAME, TXDELAY,
                        P, SLOTTIME, TXTAIL, FULLDUPLEX, SETHARDWARE,
                        DEFAULT_KISS_CONFIG_VALUES, KISS_ON, KISS_OFF,
                        NMEA_HEADER, UI_PROTOCOL_ID)

from .exceptions import SocketClosetError  # NOQA

from .util import (escape_special_codes, recover_special_codes, extract_ui,  # NOQA
                   strip_df_start, strip_nmea)

from .classes import KISS, TCPKISS, SerialKISS, TCPServerKISS  # NOQA


__author__ = 'Greg Albrecht W2GMD <oss@undef.net>'  # NOQA pylint: disable=R0801
__copyright__ = 'Copyright 2017 Greg Albrecht and Contributors'  # NOQA pylint: disable=R0801
__license__ = 'Apache License, Version 2.0'  # NOQA pylint: disable=R0801



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Setup for the Python KISS Module.

Source:: https://github.com/theskorm/kiss-fix
"""

import os
import setuptools
import sys

__title__ = 'kissfix'
__version__ = '7.0.11'
__author__ = 'Greg Albrecht W2GMD <oss@undef.net>'  # NOQA pylint: disable=R0801
__copyright__ = 'Copyright 2017 Greg Albrecht and Contributors'  # NOQA pylint: disable=R0801
__license__ = 'Apache License, Version 2.0'  # NOQA pylint: disable=R0801


def publish():
    """Function for publishing package to pypi."""
    if sys.argv[-1] == 'publish':
        os.system('python setup.py sdist')
        os.system('twine upload dist/*')
        sys.exit()


publish()


setuptools.setup(
    name=__title__,
    version=__version__,
    description='Python KISS Module.',
    author='Greg Albrecht - patches by Michael Wheeler',
    author_email='oss@undef.net',
    packages=['kissfix'],
    package_data={'': ['LICENSE']},
    package_dir={'kissfix': 'kiss'},
    license=open('LICENSE').read(),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/xssfox/kiss-fix',
    zip_safe=False,
    setup_requires=[
        'coverage >= 4.4.1',
        'nose >= 1.3.7',
        'dummyserial >= 1.0.0',
        'aprs > 6.9',
        'mocket >= 1.8.2'
    ],
    install_requires=['pyserial >= 3.4'],
    classifiers=[
        'Topic :: Communications :: Ham Radio',
        'Programming Language :: Python',
        'License :: OSI Approved :: Apache Software License'
    ],
    keywords=[
        'Ham Radio', 'APRS', 'KISS'
    ]
)
